{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRichEditToHtml.PAS, released on 2001-02-28.

The Initial Developer of the Original Code is Sbastien Buysse [sbuysse att buypin dott com]
Portions created by Sbastien Buysse are Copyright (C) 2001 Sbastien Buysse.
All Rights Reserved.

Contributor(s): Michael Beck [mbeck att bigfoot dott com],
                Andreas Hausladen [Andreas dott Hausladen att gmx dott de].

You may retrieve the latest version of this file at the Project JEDI's JVCL home page,
located at http://jvcl.delphi-jedi.org

Known Issues:

Changes (wp):
  Ported to Lazarus where TRichEdit is not available --> adapted to TRichMemo
  and renamed accordingly.

-----------------------------------------------------------------------------}
// $Id$

unit RichMemoToHTML;

interface

uses
  LCLIntf, SysUtils, Classes, Graphics, Forms,
  RichMemo;

type
  TRichMemoParagraphAttributes = record
    Alignment: TParaAlignment;
    Numbering: TParaNumbering;
  end;

  TRichMemoToHTML = class(TComponent)
  private
    FEndSection: string;
    FEndPara: string;
    FTitle: string;
    FFooter: TStrings;
    FHeader: TStrings;
    function AttToHtml(const Value: TFontParams): String;
    function ParaToHtml(Value: TRichMemoParagraphAttributes): String;
    procedure SetFooter(const Value: TStrings);
    procedure SetHeader(const Value: TStrings);
    function IsFooterStored: Boolean;
    function IsHeaderStored: Boolean;
    procedure WriteEmptyStrings(Writer: TWriter);
  protected
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ConvertToHtml(Value: TRichMemo; const FileName: String); overload;
    procedure ConvertToHtmlStrings(Value: TRichMemo; Strings: TStrings); overload;
  published
    property Title: string read FTitle write FTitle;
    property Header: TStrings read FHeader write SetHeader stored IsHeaderStored;
    property Footer: TStrings read FFooter write SetFooter stored IsFooterStored;
  end;

procedure ExportRichMemoToHtml(ARichMemo: TRichMemo; AFileName: String);

implementation

const
  // (rom) needs renaming?
//  cHTMLHeadBegin = '<HTML>';
//  cHTMLBodyBegin = '<BODY>';
//  cHTMLBodyEnd = '</BODY>';
//  cHTMLEnd = '</HTML>';
//  cHTMLTitleFmt = '<TITLE>%s</TITLE>';

  cHTMLBR = '<BR>';
//  cHTMLFontColorBegin = '<FONT COLOR=#';
//  cHTMLSize = ' SIZE=';
//  cHTMLFace = ' FACE="';
  cHTMLFontEnd = '</SPAN>';

  cHTMLBoldBegin = '<B>';
  cHTMLBoldEnd = '</B>';

  cHTMLItalicBegin = '<I>';
  cHTMLItalicEnd = '</I>';

  cHTMLStrikeoutBegin = '<STRIKE>';
  cHTMLStrikeoutEnd = '</STRIKE>';

  cHTMLUnderlineBegin = '<U>';
  cHTMLUnderlineEnd = '</U>';

  cHTMLSubscriptBegin = '<SUB>';
  cHTMLSubscriptEnd = '</SUB>';

  cHTMLSuperscriptBegin = '<SUP>';
  cHTMLSuperscriptEND = '</SUP>';

  cHTMLParaEnd = '</P>';
  cHTMLParaLeft = '<P ALIGN="LEFT">';
  cHTMLParaRight = '<P ALIGN="RIGHT">';
  cHTMLParaCenter = '<P ALIGN="CENTER">';

  cHTMLListBegin = '<LI>';
  cHTMLListEnd = '</LI>';

  cDefaultHeader: array[0..5] of string = (
    '<HTML>',
    '  <HEAD>',
    '    <TITLE><#TITLE></TITLE>',
    '    <META CHARSET="utf-8">',
    '  </HEAD>',
    '  <BODY>'
  );
  cDefaultFooter: array[0..1] of string = (
    '  </BODY>',
    '</HTML>'
  );

function ColorToHTMLColor(const Value: TColor): String;
var
  clr: TColor;
begin
  clr := ColorToRGB(Value);
  Result := Format('%.2x%.2x%.2x', [GetRValue(Clr), GetGValue(Clr), GetBValue(Clr)]);
end;


procedure ExportRichMemoToHtml(ARichMemo: TRichMemo; AFileName: String);
var
  exporter: TRichMemoToHtml;
begin
  exporter := TRichMemoToHtml.Create(nil);
  try
    exporter.ConvertToHtml(ARichMemo, AFileName);
  finally
    exporter.Free;
  end;
end;


//=== { TRichMemoToHTML } ==================================================

constructor TRichMemoToHTML.Create(AOwner: TComponent);
var
  I: Integer;
begin
  inherited Create(AOwner);
  FHeader := TStringList.Create;
  for I := 0 to High(cDefaultHeader) do
    FHeader.Add(cDefaultHeader[I]);
  FFooter := TStringList.Create;
  for I := 0 to High(cDefaultFooter) do
    FFooter.Add(cDefaultFooter[I]);
end;

destructor TRichMemoToHTML.Destroy;
begin
  FHeader.Free;
  FFooter.Free;
  inherited Destroy;
end;

function TRichMemoToHTML.AttToHtml(const Value: TFontParams): String;
var
  Size: Integer;
  bkClr: String;
begin
  FEndSection := cHTMLFontEnd;

  Size := Abs(Value.Size);
  if Size = 0 then
    Size := 10;

  if Value.HasBkClr then
    bkClr := Format('background-color: %s; ', [ColorToHTMLColor(Value.BkColor)])
  else
    bkClr := '';

  Result := Format('<SPAN style="color: #%s;%s font-size: %dpt; font-family: %s;">',
    [ColorToHTMLColor(Value.Color), bkClr, Size, Value.Name]);

  if fsBold in Value.Style then
  begin
    FEndSection := cHTMLBoldEnd + FEndSection;
    Result := Result + cHTMLBoldBegin;
  end;
  if fsItalic in Value.Style then
  begin
    FEndSection := cHTMLItalicEnd + FEndSection;
    Result := Result + cHTMLItalicBegin;
  end;
  if fsStrikeout in Value.Style then
  begin
    FEndSection := cHTMLStrikeoutEnd + FEndSection;
    Result := Result + cHTMLStrikeoutBegin;
  end;
  if fsUnderline in Value.Style then
  begin
    FEndSection := cHTMLUnderlineEnd + FEndSection;
    Result := Result + cHTMLUnderlineBegin;
  end;
  if (Value.VScriptPos = vpSubscript) then
  begin
    FEndSection := cHTMLSubscriptEnd + FEndSection;
    Result := Result + cHTMLSubscriptBegin;
  end;
  if (Value.VScriptPos = vpSuperscript) then
  begin
    FEndSection := cHTMLSuperscriptEnd + FEndSection;
    Result := Result + cHTMLSuperscriptBegin;
  end;

//  if Value.Link then
//  begin
//    FEndSection := '</a>' + FEndSection;
//    Result := Result + '<a href="#">';
//  end;
end;

function Diff(One, Two: TFontParams): Boolean;
begin
  Result := (One.Color <> Two.Color) or (One.Style <> Two.Style) or
    (One.Name <> Two.Name) or (One.Size <> Two.Size) or (One.VScriptPos <> Two.VScriptPos);
  // or (One.Link <> Two.Link);
end;

function DiffNumbering(One, Two: TParaNumbering): Boolean;
begin
  Result := (One.Style <> Two.Style) or (One.Indent <> Two.Indent) or
    (One.CustomChar <> Two.CustomChar) or (One.NumberStart <> Two.NumberStart) or
    (One.SepChar <> Two.SepChar) or (One.ForceNewNum <> Two.ForceNewNum);
end;

function DiffPara(One, Two: TRichMemoParagraphAttributes): Boolean;
begin
  Result := (One.Alignment <> Two.Alignment) or DiffNumbering(One.Numbering, Two.Numbering);
end;

procedure TRichMemoToHTML.ConvertToHtml(Value: TRichMemo; const FileName: string);
var
  S: TStringList;
begin
  S := TStringList.Create;
  try
    ConvertToHtmlStrings(Value, S);
    S.SaveToFile(FileName);
  finally
    S.Free;
  end;
end;

procedure TRichMemoToHTML.ConvertToHtmlStrings(Value: TRichMemo; Strings: TStrings);
var
  I, J: Integer;
  DefFontParams, fontParams, CurrFontParams: TFontParams;
  DPara, Para, CurrPara: TRichMemoParagraphAttributes;
  St: TStringBuilder;
  savedSelStart: Integer;
  savedSelLength: Integer;
  savedOnChange: TNotifyEvent;
  savedOnSelectionChange: TNotifyEvent;
  Text: string;
  Len: Integer;
  PreviousChar: Char;
  paraNumbering: TParaNumbering;
begin
  Para := Default(TRichMemoParagraphAttributes);
  DefFontParams := Default(TFontParams);
  fontparams := Default(TFontParams);
  CurrFontParams := Default(TFontParams);
  paraNumbering := Default(TParaNumbering);

  savedSelStart := Value.SelStart;
  savedSelLength := Value.SelLength;
  savedOnChange := Value.OnChange;
  savedOnSelectionChange := Value.OnSelectionChange;
  Strings.BeginUpdate;
  Value.Lines.BeginUpdate;
  try
    Value.OnChange := nil;
    Value.OnSelectionChange := nil;

    Strings.Clear;
    if Header.Count > 0 then
      Strings.Add(StringReplace(Header.Text, '<#TITLE>', Title, [rfReplaceAll]));

    DPara.Alignment := paLeft;
    DPara.Numbering.Style := pnNone;
    CurrPara := DPara;
    Strings.Add(ParaToHtml(CurrPara));

    DefFontParams := GetFontParams(Value.Font);
    Strings.Add(AttToHtml(DefFontParams));
    CurrFontParams := DefFontParams;

    Value.SelStart := 0;
    Value.SelectAll;
    Text := Value.SelText;
    Len := Length(Text);
    St := TStringBuilder.Create;
    try
      I := 1;
      Value.SelLength := 1;
      while I <= Len do
      begin
        // new line
        Value.SelStart := I - 1;
        Value.GetTextAttributes(Value.SelStart, fontParams);
        Value.GetParaAlignment(Value.SelStart, Para.Alignment);
        Value.GetParaNumbering(Value.SelStart, paraNumbering);
        Para.Numbering.Style := paraNumbering.Style;

        St.Length := 0;
        if DiffPara(Para, CurrPara) or (Para.Numbering.Style = pnBullet) then
        begin
          St.Append(FEndSection).Append(FEndPara);
          CurrPara.Alignment := Para.Alignment;
          CurrPara.Numbering := Para.Numbering;
          CurrFontParams := fontParams;
          St.Append(ParaToHtml(Para)).Append(AttToHtml(fontParams));
        end;

        J := I;
        PreviousChar := #0;
        while (J <= Len) and not CharInSet(Text[J], [#$A, #$B, #$D]) do { RICHEDIT uses #$B also for line breaking }
        begin
          Value.GetTextAttributes(Value.SelStart, fontParams);
          if Diff(fontParams, CurrFontParams) then
          begin
            St.Append(FEndSection);
            CurrFontParams := fontParams;
            St.Append(AttToHtml(fontParams));
            Inc(J);
            Value.SelStart := J-1;
          end
          else
          begin
            if (Text[J] = ' ') and (PreviousChar = ' ') then
              St.Append('&nbsp;')
            else
              St.Append(Text[J]);
//              St.Append(CharToHtml(Text[J]));   // FIX ME
            PreviousChar := Text[J];
            Inc(J);
            Value.SelStart := J-1;
          end;
        end;
        if I = 1 then
          Strings.Add(St.ToString())
        else
          Strings.Add(cHTMLBR + St.ToString());
        I := J + 1;
      end;
    finally
      St.Free;
    end;
    Strings.Add(FEndSection);
    Strings.Add(FEndPara);
    Strings.AddStrings(Footer);
  finally
    Value.SelStart := savedSelStart;
    Value.SelLength := savedSelLength;
    Value.OnChange := savedOnChange;
    Value.OnSelectionChange := savedOnSelectionChange;
    Strings.EndUpdate;
    Value.Lines.EndUpdate;
  end;
end;

function TRichMemoToHTML.ParaToHtml(Value: TRichMemoParagraphAttributes): string;
begin
  case Value.Alignment of
    paLeft:
      Result := 'STYLE="text-align: left;"';
    paRight:
      Result := 'STYLE="text-align: right;"';
    paCenter:
      Result := 'STYLE="text-align: center;"';
    paJustify:
      Result := 'STYLE="text-align: justify;"';
  end;
  if Value.Numbering.Style = pnBullet then
  begin
    Result := '<LI ' + Result + '>';
    FEndPara := '</LI>';
  end
  else
  begin
    Result := '<P ' + Result + '>';
    FEndPara := '</P>';
  end
end;

procedure TRichMemoToHTML.SetFooter(const Value: TStrings);
begin
  if Value <> FFooter then
    FFooter.Assign(Value);
end;

procedure TRichMemoToHTML.SetHeader(const Value: TStrings);
begin
  if Value <> FHeader then
    FHeader.Assign(Value);
end;

function TRichMemoToHTML.IsFooterStored: Boolean;
var
  I: Integer;
begin
  Result := Footer.Count <> Length(cDefaultFooter);
  if not Result then
  begin
    Result := True;
    for I := 0 to High(cDefaultFooter) do
      if Footer[I] <> cDefaultFooter[I] then
        Exit;
    Result := False;
  end;
end;

function TRichMemoToHTML.IsHeaderStored: Boolean;
var
  I: Integer;
begin
  Result := Header.Count <> Length(cDefaultHeader);
  if not Result then
  begin
    Result := True;
    for I := 0 to High(cDefaultHeader) do
      if Header[I] <> cDefaultHeader[I] then
        Exit;
    Result := False;
  end;
end;

procedure TRichMemoToHTML.WriteEmptyStrings(Writer: TWriter);
begin
  Writer.WriteListBegin;
  Writer.WriteListEnd;
end;

procedure TRichMemoToHTML.DefineProperties(Filer: TFiler);

  function DoWriteHeader: Boolean;
  begin
    Result := Header.Count = 0;
    if Result and (Filer.Ancestor <> nil) then
    begin
      Result := True;
      if Filer.Ancestor is TRichMemoToHTML then
        Result := not Header.Equals(TRichMemoToHTML(Filer.Ancestor).Header)
    end;
  end;

  function DoWriteFooter: Boolean;
  begin
    Result := Footer.Count = 0;
    if Result and (Filer.Ancestor <> nil) then
    begin
      Result := True;
      if Filer.Ancestor is TRichMemoToHTML then
        Result := not Footer.Equals(TRichMemoToHTML(Filer.Ancestor).Footer)
    end;
  end;

begin
  inherited DefineProperties(Filer);
  { Write empty Header/Footer to DFM because the default value differs from '' }
  if Filer is TWriter then
  begin
    Filer.DefineProperty('Header.Strings', nil, @WriteEmptyStrings, DoWriteHeader);
    Filer.DefineProperty('Footer.Strings', nil, @WriteEmptyStrings, DoWriteFooter);
  end;
end;

end.
